from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''performs a sort on the current result-set

EXAMPLE:
    results = [4, 12, 0, 8]
    -> sort
    result = [0, 4, 8 ,12]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('sort',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def sort(addresses):
    addresses.sort()
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return sort(addresses)



from fa import utils, context

try:
    import idautils
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('max-xrefs',
                                   description='get the result with'
                                               ' most xrefs pointing '
                                               'at it')
    return p


@context.ida_context
def max_xrefs(addresses):
    xrefs = []
    for address in addresses:
        xrefs.append((address, len([ref.frm for ref in
                                    idautils.XrefsTo(address)])))

    if len(xrefs) > 0:
        address, _ = max(xrefs, key=lambda x: x[1])
        return [address]

    return []


def run(segments, args, addresses, interpreter=None, **kwargs):
    return max_xrefs(addresses)



from collections import namedtuple
import pkg_resources
import subprocess
import tempfile
import traceback
import binascii
import sys
import re
import os

import rpyc
from rpyc import OneShotServer

sys.path.append('.')  # noqa: E402

import hjson
import click

from ida_kernwin import Form
import ida_segment
import ida_kernwin
import ida_typeinf
import ida_struct
import ida_bytes
import idautils
import ida_auto
import ida_pro
import idaapi
import idc

from fa import fainterp, fa_types

# Filename for the temporary created signature
TEMP_SIG_FILENAME = os.path.join(tempfile.gettempdir(), 'fa_tmp_sig.sig')

# IDA fa plugin filename
PLUGIN_FILENAME = 'fa_ida_plugin.py'


def open_file(filename):
    """
    Attempt to open the given filename by OS' default editor
    :param filename: filename to open
    :return: None
    """
    if sys.platform == "win32":
        try:
            os.startfile(filename)
        except Exception as error_code:
            if error_code[0] == 1155:
                os.spawnl(os.P_NOWAIT,
                          os.path.join(os.environ['WINDIR'],
                                       'system32', 'Rundll32.exe'),
                          'Rundll32.exe SHELL32.DLL, OpenAs_RunDLL {}'
                          .format(filename))
            else:
                print("other error")
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])


class IdaLoader(fainterp.FaInterp):
    """
    IDA loader
    Includes improved GUI interaction for accessing the interpreter
    functionality.
    """

    def __init__(self):
        super(IdaLoader, self).__init__()
        self._create_template_symbol = eval(self.config_get(
            'global',
            'create_symbol_template',
            'False'
        ))

    def set_symbol_template(self, status):
        """
        Should the create-temp-signature feature attempt to create a default
        signature by predefined template?
        :param status: new boolean setting
        :return: None
        """
        self._create_template_symbol = status
        self.config_set('global', 'create_symbol_template', str(status))

    def create_symbol(self):
        """
        Create a temporary symbol signature from the current function on the
        IDA screen.
        """
        self.log('creating temporary signature')

        current_ea = idc.get_screen_ea()

        signature = {
            'name': idc.get_func_name(current_ea),
            'instructions': []
        }

        if self._create_template_symbol:
            find_bytes_ida = "find-bytes-ida '"

            for ea in idautils.FuncItems(current_ea):
                mnem = idc.print_insn_mnem(ea).lower()
                opcode_size = idc.get_item_size(ea)

                # ppc
                if mnem.startswith('b') or mnem in ('lis', 'lwz', 'addi'):
                    # relative opcodes
                    find_bytes_ida += '?? ' * opcode_size
                    continue

                # arm
                if mnem.startswith('b') or mnem in ('ldr', 'str'):
                    # relative opcodes
                    find_bytes_ida += '?? ' * opcode_size
                    continue

                opcode = binascii.hexlify(idc.get_bytes(ea, opcode_size))
                formatted_hex = ' '.join(opcode[i:i + 2] for i in
                                         range(0, len(opcode), 2))
                find_bytes_ida += formatted_hex + ' '

            find_bytes_ida += "'"

            signature['instructions'].append(find_bytes_ida)
            signature['instructions'].append('function-start')
            signature['instructions'].append('set-name "{}"'.format(
                idc.get_func_name(current_ea)))

        with open(TEMP_SIG_FILENAME, 'w') as f:
            hjson.dump(signature, f, indent=4)

        self.log('Signature created at {}'.format(TEMP_SIG_FILENAME))
        return TEMP_SIG_FILENAME

    def extended_create_symbol(self):
        """
        Creates a temporary symbol of the currently active function
        and open it using OS default editor
        :return: None
        """
        filename = self.create_symbol()
        open_file(filename)

    def find_symbol(self):
        """
        Find the last create symbol signature.
        :return: None
        """
        with open(TEMP_SIG_FILENAME) as f:
            sig = hjson.load(f)

        results = self.find_from_sig_json(sig, decremental=False)

        for address in results:
            self.log('Search result: 0x{:x}'.format(address))
        self.log('Search done')

        if len(results) == 1:
            # if remote sig has a proper name, but current one is not
            ida_kernwin.jumpto(results[0])

    def verify_project(self):
        """
        Verify a valid project is currently active.
        Show IDA warning if not.
        :return: None
        """
        try:
            super(IdaLoader, self).verify_project()
        except IOError as e:
            ida_kernwin.warning(e.message)
            raise e

    def prompt_save_signature(self):
        """
        Save last-created-temp-signature if user agrees to in IDA prompt
        :return: None
        """
        self.verify_project()

        if ida_kernwin.ask_yn(1, 'Are you sure you want '
                                 'to save this signature?') != 1:
            return

        self.save_signature(TEMP_SIG_FILENAME)

    def find(self, symbol_name, use_cache=False):
        """
        Find symbol by name (as specified in SIG file)
        Show an IDA waitbox while doing so
        :param symbol_name: symbol name
        :return: output address list
        """
        ida_kernwin.replace_wait_box('Searching symbol: \'{}\'...'
                                     .format(symbol_name))
        return super(IdaLoader, self).find(symbol_name, use_cache=use_cache)

    def get_python_symbols(self, file_name=None):
        """
        Run all python scripts inside the currently active project.
        Show an IDA waitbox while doing so
        :param file_name: filter a specific filename to execute
        :return: dictionary of all found symbols
        """
        ida_kernwin.replace_wait_box('Running python scripts...')
        return super(IdaLoader, self).get_python_symbols(file_name=file_name)

    @staticmethod
    def extract_all_user_names(filename=None):
        """
        Get all user-named labels inside IDA. Also prints into output window.
        :return: dictionary of all user named labels: label_name -> ea
        """
        results = {}
        output = ''

        for ea, name in idautils.Names():
            if ida_kernwin.user_cancelled():
                return results

            if '_' in name:
                if name.split('_')[0] in ('def', 'sub', 'loc', 'jpt', 'j',
                                          'nullsub'):
                    continue
            flags = ida_bytes.get_full_flags(ea)
            if ida_bytes.has_user_name(flags):
                results[name] = ea
                output += '{} = 0x{:08x};\n'.format(name, ea)

        if filename is not None:
            with open(filename, 'w') as f:
                f.write(output)

        return results

    def set_const(self, name, value):
        super(IdaLoader, self).set_const(name, value)
        fa_types.add_const(name, value)

    def set_symbol(self, symbol_name, value):
        super(IdaLoader, self).set_symbol(symbol_name, value)
        idc.set_name(value, symbol_name, idc.SN_CHECK)

    def symbols(self, output_file_path=None):
        """
        Run find for all SIG files in currently active project.
        Show an IDA waitbox while doing so
        :param output_file_path: optional, save found symbols into output file
        :return: dictionary of found symbols
        """
        self.verify_project()
        results = {}

        try:
            ida_kernwin.show_wait_box('Searching...')
            results = super(IdaLoader, self).symbols()

            ida_kernwin.replace_wait_box('Extracting...')
            ida_symbols = IdaLoader.extract_all_user_names(output_file_path)

            results.update(ida_symbols)

        except Exception:
            traceback.print_exc()
        finally:
            ida_kernwin.hide_wait_box()

        return results

    def export(self):
        """
        Show an export dialog to export symbols and header file for given
        IDB.
        :return: None
        """

        class ExportForm(Form):
            def __init__(self):
                description = '''
                <h2>Export</h2>

                Select a directory to export IDB data into.
                '''

                Form.__init__(self,
                              r"""BUTTON YES* Save
                              Export
                              {StringLabel}
                              <#Symbols#Symbols filename:{iSymbolsFilename}>
                              <#C Header#C Header filename:{iHeaderFilename}>
                              <#ifdef_macro#ifdef'ed:{iIfdef}>
                              <#Select dir#Browse for dir:{iDir}>
                              """, {
                                  'iDir': Form.DirInput(),
                                  'StringLabel':
                                      Form.StringLabel(description,
                                                       tp=Form.FT_HTML_LABEL),
                                  'iSymbolsFilename': Form.StringInput(
                                      value='symbols.txt'),
                                  'iHeaderFilename': Form.StringInput(
                                      value='fa_structs.h'),
                                  'iIfdef': Form.StringInput(
                                      value='FA_STRUCTS_H'),
                              })
                self.__n = 0

            def OnFormChange(self, fid):
                return 1

        form = ExportForm()
        form, args = form.Compile()
        ok = form.Execute()
        if ok == 1:
            # save symbols
            symbols_filename = os.path.join(form.iDir.value,
                                            form.iSymbolsFilename.value)
            with open(symbols_filename, 'w') as f:
                results = IdaLoader.extract_all_user_names(None)
                for k, v in sorted(results.items()):
                    f.write('{} = 0x{:08x};\n'.format(k, v))

            # save c header
            idati = ida_typeinf.get_idati()
            c_header_filename = os.path.join(form.iDir.value,
                                             form.iHeaderFilename.value)

            consts_ordinal = None
            ordinals = []
            for ordinal in range(1, ida_typeinf.get_ordinal_qty(idati) + 1):
                ti = ida_typeinf.tinfo_t()
                if ti.get_numbered_type(idati, ordinal):
                    if ti.get_type_name() == 'FA_CONSTS':
                        # convert into macro definitions
                        consts_ordinal = ordinal
                    elif ti.get_type_name() in ('__va_list_tag',
                                                'va_list'):
                        continue
                    elif '$' in ti.get_type_name():
                        # skip deleted types
                        continue
                    else:
                        ordinals.append(str(ordinal))

            with open(c_header_filename, 'w') as f:
                ifdef_name = form.iIfdef.value.strip()

                if len(ifdef_name) > 0:
                    f.write('#ifndef {ifdef_name}\n'
                            '#define {ifdef_name}\n\n'
                            .format(ifdef_name=ifdef_name))

                if consts_ordinal is not None:
                    consts = re.findall('\\s*(.+?) = (.+?),',
                                        idc.print_decls(
                                            str(consts_ordinal), 0))
                    for k, v in consts:
                        f.write('#define {} ({})\n'.format(k, v))

                    # ida exports using this type
                    f.write('#define _BYTE char\n')
                    f.write('\n')

                structs_buf = idc.print_decls(','.join(ordinals),
                                              idc.PDF_DEF_BASE)

                for struct_type, struct_name in re.findall(
                        r'(struct|enum) .*?([a-zA-Z0-9_\-]+?)\s+\{',
                        structs_buf):
                    f.write(
                        'typedef {struct_type} {struct_name} {struct_name};\n'
                            .format(struct_type=struct_type,
                                    struct_name=struct_name))

                structs_buf = structs_buf.replace('__fastcall', '')
                f.write('\n')
                f.write(structs_buf)
                f.write('\n')

                if len(ifdef_name) > 0:
                    f.write('#endif // {ifdef_name}\n'
                            .format(ifdef_name=ifdef_name))

        form.Free()

    def set_input(self, input_):
        """
        Mock for change_input. Just reload current loaded data settings.
        :param input_: doesn't matter
        :return: None
        """
        self.endianity = '>' if idaapi.get_inf_structure().is_be() else '<'
        self._input = input_
        self.reload_segments()

    def reload_segments(self):
        """
        memory searches will use IDA's API instead
        which is much faster so this is just a stub.
        :return: None
        """
        return

    def interactive_settings(self):
        """
        Show settings dialog
        :return: None
        """

        class SettingsForm(Form):
            def __init__(self, signatures_root, use_template):
                description = '''
                <h2>Settings</h2>
                <div>
                Here you can change global FA settings.
                </div>
                <div>
                    <a href="https://github.com/doronz88/fa">
                    For more info</a>
                </div>
                '''

                Form.__init__(self,
                              r"""BUTTON YES* Save
                              FA Settings
                              {{FormChangeCb}}
                              {{StringLabel}}
                              <Signatures root :{{signaturesRoot}}>
                              <Temp signature generation :{{signatureGeneration}}>
                              """.format(signatures_root), {
                                  'FormChangeCb':
                                      Form.FormChangeCb(self.OnFormChange),
                                  'signaturesRoot':
                                      Form.DirInput(value=signatures_root),
                                  'StringLabel':
                                      Form.StringLabel(description,
                                                       tp=Form.FT_HTML_LABEL),
                                  'signatureGeneration':
                                      Form.DropdownListControl(
                                          items=['Default',
                                                 'Using function bytes'],
                                          readonly=True,
                                          selval=use_template),
                              })
                self.__n = 0

            def OnFormChange(self, fid):
                return 1

        f = SettingsForm(self._signatures_root, self._create_template_symbol)
        f, args = f.Compile()
        ok = f.Execute()
        if ok == 1:
            self.set_signatures_root(f.signaturesRoot.value, save=True)
            self.set_symbol_template(f.signatureGeneration.value == 1)
        f.Free()

    def interactive_set_project(self):
        """
        Show set-project dialog
        :return: None
        """

        class SetProjectForm(Form):
            def __init__(self, signatures_root, projects, current):
                description = '''
                <h2>Project Selector</h2>
                <div>
                Select project you wish to work on from your
                signatures root:
                </div>
                <div><pre>{}</pre></div>
                <div><i>(Note: You may change this in config.ini)</i></div>
                <div>
                    <a href="https://github.com/doronz88/fa#projects">
                    For more info</a>
                </div>
                '''.format(signatures_root)

                Form.__init__(self,
                              r"""BUTTON YES* OK
                              FA Project Select
                              {{FormChangeCb}}
                              {{StringLabel}}
                              <Set Project :{{cbReadonly}}>
                              """.format(signatures_root), {
                                  'FormChangeCb':
                                      Form.FormChangeCb(self.OnFormChange),
                                  'cbReadonly':
                                      Form.DropdownListControl(
                                          items=projects,
                                          readonly=True,
                                          selval=projects.index(current)
                                          if current in projects else 0),
                                  'StringLabel':
                                      Form.StringLabel(description,
                                                       tp=Form.FT_HTML_LABEL),
                              })
                self.__n = 0

            def OnFormChange(self, fid):
                return 1

        projects = self.list_projects()
        f = SetProjectForm(self._signatures_root, projects, self._project)
        f, args = f.Compile()
        ok = f.Execute()
        if ok == 1:
            self.set_project(projects[f.cbReadonly.value])
        f.Free()


fa_instance = None

Action = namedtuple('action', 'name icon_filename handler label hotkey')


def add_action(action):
    """
    Add an ida-action
    :param action: action given as the `Action` namedtuple
    :return: None
    """

    class Handler(ida_kernwin.action_handler_t):
        def __init__(self):
            ida_kernwin.action_handler_t.__init__(self)

        def activate(self, ctx):
            action.handler()
            return 1

        def update(self, ctx):
            return ida_kernwin.AST_ENABLE_FOR_WIDGET

    act_icon = -1
    if action.icon_filename:
        icon_full_filename = \
            pkg_resources.resource_filename('fa',
                                            os.path.join(
                                                'res',
                                                'icons',
                                                action.icon_filename))
        with open(icon_full_filename, 'rb') as f:
            icon_data = f.read()
        act_icon = ida_kernwin.load_custom_icon(data=icon_data, format="png")

    act_name = action.name

    ida_kernwin.unregister_action(act_name)
    if ida_kernwin.register_action(ida_kernwin.action_desc_t(
            act_name,  # Name. Acts as an ID. Must be unique.
            action.label,  # Label. That's what users see.
            Handler(),  # Handler. Called when activated, and for updating
            action.hotkey,  # Shortcut (optional)
            None,  # Tooltip (optional)
            act_icon)):  # Icon ID (optional)

        # Insert the action in the menu
        if not ida_kernwin.attach_action_to_menu(
                "FA/", act_name, ida_kernwin.SETMENU_APP):
            print("Failed attaching to menu.")

        # Insert the action in a toolbar
        if not ida_kernwin.attach_action_to_toolbar("fa", act_name):
            print("Failed attaching to toolbar.")

        class Hooks(ida_kernwin.UI_Hooks):
            def finish_populating_widget_popup(self, widget, popup):
                if ida_kernwin.get_widget_type(widget) == \
                        ida_kernwin.BWN_DISASM:
                    ida_kernwin.attach_action_to_popup(widget,
                                                       popup,
                                                       act_name,
                                                       None)

        hooks = Hooks()
        hooks.hook()


def load_ui():
    """
    Load FA's GUI buttons
    :return: None
    """
    actions = [
        Action(name='fa:settings',
               icon_filename='settings.png',
               handler=fa_instance.interactive_settings,
               label='Settings',
               hotkey=None),

        Action(name='fa:set-project',
               icon_filename='suitcase.png',
               handler=fa_instance.interactive_set_project,
               label='Set project...',
               hotkey='Ctrl+6'),

        Action(name='fa:symbols', icon_filename='find_all.png',
               handler=fa_instance.symbols,
               label='Find all project\'s symbols',
               hotkey='Ctrl+7'),

        Action(name='fa:export', icon_filename='export.png',
               handler=fa_instance.export,
               label='Export symbols',
               hotkey=None),

        Action(name='fa:extended-create-signature',
               icon_filename='create_sig.png',
               handler=fa_instance.extended_create_symbol,
               label='Create temp signature...',
               hotkey='Ctrl+8'),

        Action(name='fa:find-symbol',
               icon_filename='find.png',
               handler=fa_instance.find_symbol,
               label='Find last created temp signature',
               hotkey='Ctrl+9'),

        Action(name='fa:prompt-save',
               icon_filename='save.png',
               handler=fa_instance.prompt_save_signature,
               label='Save last created temp signature',
               hotkey='Ctrl+0'),
    ]

    # init toolbar
    ida_kernwin.delete_toolbar('fa')
    ida_kernwin.create_toolbar('fa', 'FA Toolbar')

    # init menu
    ida_kernwin.delete_menu('fa')
    ida_kernwin.create_menu('fa', 'FA')

    for action in actions:
        add_action(action)


def install():
    """
    Install FA ida plugin
    :return: None
    """
    fa_plugin_dir = os.path.join(
        idaapi.get_user_idadir(), 'plugins')

    if not os.path.exists(fa_plugin_dir):
        os.makedirs(fa_plugin_dir)

    fa_plugin_filename = os.path.join(fa_plugin_dir, PLUGIN_FILENAME)
    if os.path.exists(fa_plugin_filename):
        IdaLoader.log('already installed')
        return

    with open(fa_plugin_filename, 'w') as f:
        f.writelines("""from __future__ import print_function
try:
    from fa.ida_plugin import PLUGIN_ENTRY, FAIDAPlugIn
except ImportError:
    print("[WARN] Could not load FA plugin. "
          "FA Python package doesn\'t seem to be installed.")
""")

    idaapi.load_plugin(PLUGIN_FILENAME)

    IdaLoader.log('Successfully installed :)')


@click.command()
@click.option('-s', '--service', type=click.IntRange(1024, 65535), help='execute in rpyc service mode at given port')
def main(service):
    plugin_main(service)


class FaService(rpyc.Service):
    ida_segment = ida_segment
    ida_kernwin = ida_kernwin
    ida_typeinf = ida_typeinf
    ida_struct = ida_struct
    ida_bytes = ida_bytes
    idautils = idautils
    ida_auto = ida_auto
    ida_pro = ida_pro
    idaapi = idaapi
    idc = idc

    @staticmethod
    def load_module(name, filename):
        return fainterp.FaInterp.get_module(name, filename)


def plugin_main(service=None):
    global fa_instance

    fa_instance = IdaLoader()
    fa_instance.set_input('ida')

    load_ui()

    IdaLoader.log('''
    ---------------------------------
    FA Loaded successfully

    Quick usage:
    print(fa_instance.find(symbol_name)) # searches for the specific symbol
    fa_instance.get_python_symbols(filename=None) # run project's python
                                                    scripts (all or single)
    fa_instance.set_symbol_template(status) # enable/disable template temp
                                              signature
    fa_instance.symbols() # searches for the symbols in the current project
    ---------------------------------
    ''')

    if service:
        t = OneShotServer(FaService, port=service, protocol_config={
            'allow_all_attrs': True,
            'allow_setattr': True,
        })
        t.start()

    # TODO: consider adding as autostart script
    # install()


try:
    class FAIDAPlugIn(idaapi.plugin_t):
        wanted_name = "FA"
        wanted_hotkey = "Shift-,"
        flags = 0
        comment = ""
        help = "Load FA in IDA Pro"

        def init(self):
            plugin_main()
            return idaapi.PLUGIN_KEEP

        def run(self, args):
            pass

        def term(self):
            pass
except TypeError:
    print('ignoring rpyc bug')


def PLUGIN_ENTRY():
    """
    Entry point for IDA plugins
    :return:
    """
    return FAIDAPlugIn()


if __name__ == '__main__':
    # Entry point for IDA in script mode (-S)
    main(standalone_mode=False, args=idc.ARGV[1:])



import pdb

from fa import utils


def get_parser():
    p = utils.ArgumentParserNoExit('trace',
                                   description='sets a pdb breakpoint')
    return p


def trace(addresses):
    pdb.set_trace()
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return trace(addresses)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''verifies the result-list contains a single value

EXAMPLE #1:
    results = [4, 12, 0, 8]
    -> verify-single
    result = []

EXAMPLE #2:
    results = [4]
    -> verify-single
    result = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('verify-single',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def verify_single(addresses):
    return addresses if len(addresses) == 1 else []


def run(segments, args, addresses, interpreter=None, **kwargs):
    return verify_single(addresses)



from argparse import RawTextHelpFormatter
try:
    # flake8: noqa
    from keystone import *
except ImportError:
    print('keystone-engine module not installed')

import binascii

from fa.commands import find_bytes
from fa import utils

DESCRIPTION = '''use keystone to search for the supplied opcodes

EXAMPLE:
    0x00000000: push {r4-r7, lr}
    0x00000004: mov r0, r1

    results = []
    -> keystone-find-opcodes --bele KS_ARCH_ARM KS_MODE_ARM 'mov r0, r1;'
    result = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('keystone-find-opcodes',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('--bele', action='store_true',
                   help='figure out the endianity from IDA instead of '
                        'explicit mode')
    p.add_argument('--or', action='store_true',
                   help='mandatory. expands search results')
    p.add_argument('arch',
                   help='keystone architecture const (evaled)')
    p.add_argument('mode',
                   help='keystone mode const (evald)')
    p.add_argument('code',
                   help='keystone architecture const (opcodes to compile)')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    arch = eval(args.arch)
    mode = eval(args.mode)

    if args.bele:
        mode |= KS_MODE_BIG_ENDIAN if \
            interpreter.endianity == '>' else KS_MODE_LITTLE_ENDIAN

    ks = Ks(arch, mode)
    compiled_buf = bytearray(ks.asm(args.code)[0])

    setattr(args, 'hex_str', binascii.hexlify(compiled_buf).decode('utf8'))
    return find_bytes.run(segments, args, addresses, **kwargs)



from fa import utils, context
from fa.commands import function_start

try:
    import idc
    import idautils
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit(prog='xrefs-to',
                                   description='search for xrefs pointing '
                                               'at given parameter')
    p.add_argument('--function-start', action='store_true',
                   help='goto function prolog for each xref')
    p.add_argument('--or', action='store_true',
                   help='expand the current result set')
    p.add_argument('--and', action='store_true',
                   help='reduce the current result set')
    p.add_argument('--name', help='parameter as label name')
    p.add_argument('--bytes', help='parameter as bytes')
    return p


@context.ida_context
def run(segments, args, addresses, interpreter=None, **kwargs):
    if args.name:
        ea = idc.LocByName(args.name)
        occurrences = [ea] if ea != idc.BADADDR else []
    else:
        occurrences = list(utils.ida_find_all(args.bytes))

    frm = set()
    for ea in occurrences:
        froms = [ref.frm for ref in idautils.XrefsTo(ea)]

        if args.function_start:
            froms = [function_start.get_function_start(segments, ea)
                     for ea in froms]

        frm.update(frm for frm in froms if frm != idc.BADADDR)

    retval = set()
    retval.update(addresses)

    if getattr(args, 'or'):
        retval.update(frm)

    elif getattr(args, 'and'):
        retval.intersection_update(frm)

    return list(retval)



from argparse import RawTextHelpFormatter
import binascii

from fa import utils

DESCRIPTION = '''reduce the result-set to those matching the given bytes

EXAMPLE:
    0x00000000: 01 02 03 04
    0x00000004: 05 06 07 08

    results = [0, 2, 4, 6, 8]
    -> verify-bytes '05 06 07 08'
    results = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('verify-bytes',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('hex_str')
    return p


def verify_bytes(addresses, hex_str, segments=None, until=None):
    magic = binascii.unhexlify(''.join(hex_str.split(' ')))

    results = [ea for ea in addresses
               if utils.read_memory(segments, ea, len(magic)) == magic]

    if len(results) > 0:
        return results

    return results


def run(segments, args, addresses, interpreter=None, **kwargs):
    until = None
    if 'until' in args and args.until is not None:
        until = args.until
    return verify_bytes(addresses, args.hex_str,
                        segments=segments, until=until)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''leave only results fitting required alignment

EXAMPLE:
    results = [0, 2, 4, 6, 8]
    -> verify-aligned 4
    results = [0, 4, 8]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('verify-aligned',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('value', type=int)
    return p


def aligned(addresses, value):
    return [ea for ea in addresses if ea % value == 0]


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(aligned(addresses, args.value))



from argparse import RawTextHelpFormatter
import binascii

from fa import utils

DESCRIPTION = '''expands the result-set with the occurrences of the given bytes

EXAMPLE:
    0x00000000: 01 02 03 04
    0x00000004: 05 06 07 08

    results = []
    -> find-bytes 01020304
    result = [0]

    -> find-bytes 05060708
    results = [0, 4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('find-bytes',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('hex_str')
    return p


def find_bytes(hex_str, segments=None):
    needle = binascii.unhexlify(''.join(hex_str.split(' ')))
    return utils.find_raw(needle, segments=segments)


def run(segments, args, addresses, interpreter=None, **kwargs):
    results = list(find_bytes(args.hex_str, segments=segments))
    return addresses + results



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''add an hard-coded value into resultset

EXAMPLE:
    results = []
    -> add 80
    result = [80]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('add',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('value')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return addresses + [eval(args.value)]



from fa.commands.locate import locate
from fa import utils, context

try:
    import idautils
    import idc
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('verify-ref',
                                   description='verifies a given reference '
                                               'exists to current result set')
    p.add_argument('--code', action='store_true',
                   default=False, help='include code references')
    p.add_argument('--data', action='store_true',
                   default=False, help='include data references')
    p.add_argument('--name', default=None, help='symbol name')
    return p


@context.ida_context
def verify_ref(addresses, name=None, code=False, data=False):
    if name is not None:
        symbol = locate(name)

        if symbol == idc.BADADDR:
            return

    for address in addresses:
        refs = []
        if code:
            refs += list(idautils.CodeRefsFrom(address, 1))
        if data:
            refs += list(idautils.DataRefsFrom(address))

        if len(refs) == 0:
            continue

        for ref in refs:
            if name is not None:
                if address + 4 != ref and symbol == ref:
                    yield address
                    break
            else:
                if address + 4 != ref:
                    yield address
                    break


@utils.yield_unique
def verify_ref_unique(addresses, name, code=False, data=False):
    for address in verify_ref(addresses, name, code=code, data=data):
        yield address


@context.ida_context
def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(set(verify_ref(addresses, args.name,
                               code=args.code, data=args.data)))



from fa import utils, context

try:
    import idautils
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('xref',
                                   description='goto xrefs pointing at'
                                               ' current search results')
    return p


@context.ida_context
def xref(addresses):
    for address in addresses:
        for ref in idautils.XrefsTo(address):
            yield ref.frm


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(xref(addresses))



from fa import utils, fa_types


def get_parser():
    p = utils.ArgumentParserNoExit('set-enum',
                                   description='define an enum value')
    p.add_argument('enum_name')
    p.add_argument('enum_key')
    return p


def set_enum(addresses, enum_name, enum_key):
    for ea in addresses:
        enum = fa_types.FaEnum(enum_name)
        enum.add_value(enum_key, ea)
        enum.update_idb()
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return set_enum(addresses, args.enum_name, args.enum_key)



from fa import utils, fa_types, context

try:
    import idc
    import ida_auto
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('set-type',
                                   description='sets the type in '
                                               'the disassembler')
    p.add_argument('type_str')
    return p


def set_type_single(address, type_):
    if isinstance(type_, fa_types.FaStruct) or \
            isinstance(type_, fa_types.FaEnum):
        type_str = type_.get_name()
    else:
        type_str = type_

    idc.SetType(address, type_str)
    ida_auto.auto_wait()


@context.ida_context
def set_type(addresses, type_):
    for ea in addresses:
        set_type_single(ea, type_)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return set_type(addresses, args.type_str)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''exit if current resultset is empty

EXAMPLE:
    results = []

    -> stop-if-empty
    add 1

    results = []
'''


def get_parser():
    p = utils.ArgumentParserNoExit('stop-if-empty',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    if len(addresses) == 0:
        # just a big enough value which is always greater
        # then max available pc
        interpreter.set_pc(0xffffffff)
    return addresses



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''go back to previous result-set saved by 'store' command.

EXAMPLE:
    results = [0, 4, 8]
    store foo

    find-bytes 12345678
    results = [0, 4, 8, 10, 20]

    -> load foo
    results = [0, 4, 8]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('load',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('name', help='name of variable in history to go back '
                                'to')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return interpreter.get_variable(args.name)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''symmetric difference between two or more variables

EXAMPLE:
    results = [0, 4, 8]
    store a
    ...
    results = [0, 12, 20]
    store b

    -> symdiff a b
    results = [4, 8, 12, 20]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('symdiff',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('variables', nargs='+', help='variable names')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    first_var = args.variables[0]
    results = set(interpreter.get_variable(first_var))

    for c in args.variables[1:]:
        results.symmetric_difference_update(interpreter.get_variable(c))

    return list(results)



import sys
from argparse import RawTextHelpFormatter
import binascii
import six

from fa.commands import find_bytes

DESCRIPTION = '''expands the result-set with the occurrences of the given
string

EXAMPLE:
    0x00000000: 01 02 03 04
    0x00000004: 05 06 07 08
    0x00000008: 30 31 32 33 -> ASCII '0123'

    results = []
    -> find-str '0123'

    result = [8]
'''


def get_parser():
    p = find_bytes.get_parser()
    p.prog = 'find-str'
    p.description = DESCRIPTION
    p.formatter_class = RawTextHelpFormatter
    p.add_argument('--null-terminated', action='store_true')
    return p


def find_str(string, null_terminated=False):
    hex_str = binascii.hexlify(string)
    if null_terminated:
        hex_str += '00'
    return find_bytes.find_bytes(hex_str)


def run(segments, args, addresses, interpreter=None, **kwargs):
    hex_str = binascii.hexlify(six.b(args.hex_str))

    if sys.version[0] == '3':
        hex_str = hex_str.decode()

    if args.null_terminated:
        hex_str += '00'
    setattr(args, 'hex_str', hex_str)
    return find_bytes.run(segments, args, addresses, **kwargs)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''perform an 'if' statement to create conditional branches
using an eval'ed expression

EXAMPLE:
    results = [0, 4, 8]

    verify-single
    store a

    # jump to a_is_single_label since a == []
    -> python-if a a_is_single_label
    set-name a_isnt_single
    b end

    label a_is_single_label
    set-name a_is_single

    label end
'''


def get_parser():
    p = utils.ArgumentParserNoExit('python-if',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('cond', help='condition to evaluate (being eval\'ed)')
    p.add_argument('label', help='label to jump to if condition is true')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    if eval(args.cond, interpreter.get_all_variables()):
        interpreter.set_pc(args.label)
        # pc is incremented by 1, after each instruction
        interpreter.dec_pc()
    return addresses



from fa import utils, context

try:
    import ida_bytes
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('make-unknown',
                                   description='convert into an unknown block')
    return p


@context.ida_context
def make_unknown(addresses):
    for ea in addresses:
        ida_bytes.del_items(ea)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return make_unknown(addresses)



from fa import utils

DESCRIPTION = '''prints the current result-set (for debugging)'''


def get_parser():
    p = utils.ArgumentParserNoExit('print',
                                   description=DESCRIPTION)
    p.add_argument('phrase', nargs='?', default='', help='optional string')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    log_line = 'FA Debug Print: {}\n'.format(args.phrase)
    for ea in addresses:
        log_line += '\t0x{:x}\n'.format(ea)
    print(log_line)
    return addresses



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idc
except ImportError:
    pass

DESCRIPTION = '''reduce the result-set to those matching the given instruction

EXAMPLE #1:
    0x00000000: mov r0, r1
    0x00000004: mov r1, r2
    0x00000008: push {r4}

    results = [0, 2, 4, 6, 8]
    -> verify-operand mov
    results = [0, 4]

EXAMPLE #2:
    0x00000000: mov r0, r1
    0x00000004: mov r1, r2
    0x00000008: push {r4}

    results = [0, 2, 4, 6, 8]
    -> verify-operand mov --op1 2
    results = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('verify-operand',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('name')
    p.add_argument('--op0')
    p.add_argument('--op1')
    p.add_argument('--op2')
    return p


@context.ida_context
@utils.yield_unique
def verify_operand(addresses, mnem, op0=None, op1=None, op2=None):
    for address in addresses:
        current_mnem = idc.print_insn_mnem(address).lower()
        if current_mnem == mnem:
            if not op0 and not op1 and not op2:
                yield address
                continue

            regs_description = []

            if op0:
                regs_description.append((0, op0))
            if op1:
                regs_description.append((1, op1))
            if op2:
                regs_description.append((2, op2))

            for description in regs_description:
                index, values = description
                if idc.get_operand_value(address, index) not in values:
                    break
            else:
                yield address


def run(segments, args, addresses, interpreter=None, **kwargs):
    op0 = [eval(i) for i in args.op0.split(',')] if args.op0 else None
    op1 = [eval(i) for i in args.op1.split(',')] if args.op1 else None
    op2 = [eval(i) for i in args.op2.split(',')] if args.op2 else None
    return list(verify_operand(addresses,
                               args.name,
                               op0=op0,
                               op1=op1,
                               op2=op2))



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''make the resultset unique

EXAMPLE:
    results = [0, 4, 8, 8, 12]
    -> unique
    result = [0, 4, 8, 12]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('unique',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(set(addresses))



from fa.utils import ArgumentParserNoExit


def get_parser():
    p = ArgumentParserNoExit('set-name',
                             description='set symbol name')
    p.add_argument('name')
    return p


def set_name(addresses, name, interpreter):
    for ea in addresses:
        interpreter.set_symbol(name, ea)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return set_name(addresses, args.name, interpreter)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''clears the current result-set

EXAMPLE:
    results = [0, 4, 8]
    -> clear
    results = []
'''


def get_parser():
    p = utils.ArgumentParserNoExit('clear',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return []



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idautils
except ImportError:
    pass

DESCRIPTION = '''goto reference

EXAMPLE:
    0x00000000: ldr r0, =0x12345678

    results = [0]
    -> goto-ref --data
    results = [0x12345678]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('goto-ref',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('--code', action='store_true',
                   default=False, help='include code references')
    p.add_argument('--data', action='store_true',
                   default=False, help='include data references')
    return p


@context.ida_context
def goto_ref(addresses, code=False, data=False):
    for address in addresses:
        refs = []
        if code:
            refs += list(idautils.CodeRefsFrom(address, 0))
        if data:
            refs += list(idautils.DataRefsFrom(address))

        if len(refs) == 0:
            continue

        for ref in refs:
            if address + 4 != ref:
                yield ref


def goto_ref_unique(addresses, code=False, data=False):
    for address in goto_ref(addresses, code=code, data=data):
        yield address


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(goto_ref_unique(addresses,
                                code=args.code,
                                data=args.data))



import argparse
import inspect
import os
import warnings

IDA_MODULE = False

try:
    import idc
    import ida_struct

    IDA_MODULE = True
except ImportError:
    pass


def index_of(needle, haystack):
    try:
        return haystack.index(needle)
    except ValueError:
        return -1


def find_raw(needle, segments=None):
    if segments is None:
        segments = dict()

    if IDA_MODULE:
        # ida optimization
        needle = bytearray(needle)
        payload = ' '.join(['{:02x}'.format(b) for b in needle])
        for address in ida_find_all(payload):
            yield address
        return

    for segment_ea, data in segments.items():
        offset = index_of(needle, data)
        extra_offset = 0

        while offset != -1:
            address = segment_ea + offset + extra_offset
            yield address

            extra_offset += offset+1
            data = data[offset+1:]

            offset = index_of(needle, data)


def ida_find_all(payload):
    ea = idc.find_binary(0, idc.SEARCH_DOWN | idc.SEARCH_REGEX, payload)
    while ea != idc.BADADDR:
        yield ea
        ea = idc.find_binary(ea + 1, idc.SEARCH_DOWN | idc.SEARCH_REGEX,
                             payload)


def read_memory(segments, ea, size):
    if IDA_MODULE:
        return idc.get_bytes(ea, size)

    for segment_ea, data in segments.items():
        if (ea <= segment_ea + len(data)) and (ea >= segment_ea):
            offset = ea - segment_ea
            return data[offset:offset+size]


def yield_unique(func):
    def wrapper(*args, **kwargs):
        results = set()
        for i in func(*args, **kwargs):
            if i not in results:
                yield i
                results.add(i)
    return wrapper


class ArgumentParserNoExit(argparse.ArgumentParser):
    def error(self, message):
        raise ValueError(message)


def deprecated(function):
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    filename = module.__file__
    command_name = os.path.splitext(os.path.basename(filename))[0]

    warnings.warn('command: "{}" is deperected and will be removed in '
                  'the future.'.format(command_name, DeprecationWarning))
    return function


def add_struct_to_idb(name):
    idc.import_type(-1, name)


def find_or_create_struct(name):
    sid = ida_struct.get_struc_id(name)
    if sid == idc.BADADDR:
        sid = idc.add_struc(-1, name, 0)
        print("added struct \"{0}\", id: {1}".format(name, sid))
    else:
        print("struct \"{0}\" already exists, id: ".format(name, sid))

    add_struct_to_idb(name)

    return sid



#!/usr/bin/python
import os
import socket
import subprocess
from collections import namedtuple

import IPython
import click
import rpyc
from termcolor import cprint

IDA_PLUGIN_PATH = os.path.abspath(os.path.join((os.path.dirname(__file__), 'ida_plugin.py')))

TerminalProgram = namedtuple('TerminalProgram', 'executable args')


def is_windows():
    return os.name == 'nt'


SUPPORTED_TERMINALS = [
    TerminalProgram(executable='kitty', args=['bash', '-c']),
    TerminalProgram(executable='gnome-terminal', args=['-x', 'bash', '-c']),
    TerminalProgram(executable='xterm', args=['-e']),
]


def get_free_port():
    s = socket.socket()
    s.bind(('', 0))
    port = s.getsockname()[1]
    s.close()
    return port


def does_program_exist(program):
    return 0 == subprocess.Popen(['which', program]).wait()


def execute_in_new_terminal(cmd):
    if is_windows():
        subprocess.Popen(cmd)
        return

    for terminal in SUPPORTED_TERMINALS:
        if does_program_exist(terminal.executable):
            subprocess.Popen([terminal.executable] + terminal.args + [' '.join(cmd)])
            return


def get_client(ida, payload, loader=None, processor_type=None, accept_defaults=False, log_file_path=None):
    port = get_free_port()
    args = [ida]

    if processor_type is not None:
        args.append('-p{}'.format(processor_type))

    if loader is not None:
        args.append('-T{}'.format(loader))

    if log_file_path is not None:
        args.append('-L{}'.format(log_file_path))

    if accept_defaults:
        args.append('-A')

    args.append('\'-S{} --service {}\''.format(IDA_PLUGIN_PATH, port))
    args.append(payload)

    execute_in_new_terminal(args)

    while True:
        try:
            client = rpyc.connect('localhost', port, config={
                # this is meant to disable the timeout
                'sync_request_timeout': None,
                'allow_all_attrs': True,
                'allow_setattr': True,
            })
            break
        except socket.error:
            pass

    return client


def launch_ida_in_service_mode(ida, payload, loader=None):
    client = get_client(ida, payload, loader)
    cprint('use `client.root` variable to access the remote object', 'cyan')
    IPython.embed()
    client.close()


@click.command()
@click.argument('ida', type=click.Path(exists=True))
@click.argument('payload', type=click.Path(exists=True))
@click.option('-l', '--loader', required=False)
def shell(ida, payload, loader):
    launch_ida_in_service_mode(ida, payload, loader)


if __name__ == '__main__':
    shell()



from abc import abstractmethod
from collections import namedtuple


try:
    import idc
    import idaapi
    import ida_auto
    import ida_enum
    import ida_struct

    IDA_MODULE = True
except ImportError:
    pass


class FaType(object):
    def __init__(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def exists(self):
        return -1 != ida_struct.get_struc_id(self._name)

    @abstractmethod
    def update_idb(self):
        pass


class FaEnum(FaType):
    def __init__(self, name):
        super(FaEnum, self).__init__(name)
        self._values = {}

    def add_value(self, name, value):
        self._values[value] = name

    def update_idb(self):
        id = ida_enum.get_enum(self._name)
        if idc.BADADDR == id:
            id = ida_enum.add_enum(idc.BADADDR, self._name, idaapi.decflag())

        keys = self._values.keys()
        keys.sort()

        for k in keys:
            ida_enum.add_enum_member(id, self._values[k], k)


class FaStruct(FaType):
    Field = namedtuple('Field', ['name', 'type', 'offset'])

    def __init__(self, name):
        super(FaStruct, self).__init__(name)
        self._fields = []

    def add_field(self, name, type_, offset=0xffffffff):
        self._fields.append(self.Field(name, type_, offset))

    def update_idb(self, delete_existing_members=True):
        sid = ida_struct.get_struc_id(self._name)
        sptr = ida_struct.get_struc(sid)

        if sid == idc.BADADDR:
            sid = ida_struct.add_struc(idc.BADADDR, self._name, 0)
            sptr = ida_struct.get_struc(sid)
        else:
            if delete_existing_members:
                ida_struct.del_struc_members(sptr, 0, 0xffffffff)

        for f in self._fields:
            ida_struct.add_struc_member(sptr, f.name, f.offset,
                                        (idc.FF_BYTE | idc.FF_DATA)
                                        & 0xFFFFFFFF,
                                        None, 1)
            member_name = "{}.{}".format(self._name, f.name)
            idc.SetType(idaapi.get_member_by_fullname(member_name)[0].id,
                        f.type)

        ida_auto.auto_wait()


def add_const(name, value):
    fa_consts = FaEnum('FA_CONSTS')
    fa_consts.add_value(name, value)
    fa_consts.update_idb()



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''save current result-set in a variable.
You can later load the result-set using 'load'

EXAMPLE:
    results = [0, 4, 8]
    -> store foo

    find-bytes --or 12345678
    results = [0, 4, 8, 10, 20]

    load foo
    results = [0, 4, 8]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('store',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('name', help='name of variable to use')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    interpreter.set_variable(args.name, addresses)
    return addresses



from argparse import RawTextHelpFormatter

try:
    import idc
except ImportError:
    pass

from fa import utils, context

DESCRIPTION = '''add comment for given addresses

EXAMPLE:
    0x00000200: 01 02 03 04
    0x00000204: 30 31 32 33

    results = [0x200]
    -> make-comment 'bla bla'
    results = [0x200]

    0x00000200: 01 02 03 04 ; bla bla
    0x00000204: 30 31 32 33
'''


@context.ida_context
def make_comment(addresses, comment):
    for ea in addresses:
        idc.set_cmt(ea, comment, 0)
        yield ea


def get_parser():
    p = utils.ArgumentParserNoExit()
    p.add_argument('comment', help='comment string')

    p.prog = 'make-comment'
    p.description = DESCRIPTION
    p.formatter_class = RawTextHelpFormatter
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(make_comment(addresses, args.comment))



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idautils
except ImportError:
    pass

DESCRIPTION = '''get all function's lines

EXAMPLE:
    0x00000000: push {r4-r7, lr} -> function's prolog
    0x00000004: mov r1, r0
    ...
    0x000000c0: mov r0, r5
    ...
    0x000000f0: push {r4-r7, pc} -> function's epilog

    results = [0xc0]
    -> function-lines
    result = [0, 4, ..., 0xc0, ..., 0xf0]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('function-lines',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    g = p.add_mutually_exclusive_group()
    g.add_argument('--after', action='store_true',
                   help='include only function lines which occur after current'
                        'resultset')
    g.add_argument('--before', action='store_true',
                   help='include only function lines which occur before '
                        'current resultset')
    return p


@context.ida_context
def function_lines(addresses, after=False, before=False):
    for address in addresses:
        for item in idautils.FuncItems(address):
            if after:
                if item > address:
                    yield item
            elif before:
                if item < address:
                    yield item
            else:
                yield item


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(function_lines(addresses, after=args.after,
                               before=args.before))



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''get the result appearing the most in the result-set

EXAMPLE:
    results = [0, 4, 4, 8, 12]
    -> most-common
    result = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('most-common',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


def most_common(addresses):
    addresses = list(addresses)
    if len(addresses) == 0:
        return []
    return [max(set(addresses), key=addresses.count)]


def run(segments, args, addresses, interpreter=None, **kwargs):
    return most_common(addresses)



from fa import utils, context

try:
    import idc
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('make-literal',
                                   description='convert into a literal')
    return p


@context.ida_context
def make_literal(addresses):
    for ea in addresses:
        idc.create_strlit(ea, idc.BADADDR)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return make_literal(addresses)



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''advance the result-set by a given offset

EXAMPLE:
    results = [0, 4, 8, 12]
    -> offset 4
    result = [4, 8, 12, 16]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('offset',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('offset')
    return p


def offset(addresses, advance_by):
    for ea in addresses:
        yield ea + advance_by


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(offset(addresses, eval(args.offset)))



from fa import utils


def get_parser():
    p = utils.ArgumentParserNoExit('run',
                                   description='run another SIG file')
    p.add_argument('name', help='SIG filename')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    interpreter.find_from_sig_path(args.name)

    # return an empty result-set
    return []



from fa import utils, fa_types


def get_parser():
    p = utils.ArgumentParserNoExit('set-struct-member',
                                   description='add a struct member')
    p.add_argument('struct_name')
    p.add_argument('member_name')
    p.add_argument('member_type')
    return p


def set_struct_member(addresses, struct_name, member_name, member_type):
    for ea in addresses:
        enum = fa_types.FaStruct(struct_name)
        enum.add_field(member_name, member_type, offset=ea)
        enum.update_idb(delete_existing_members=False)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return set_struct_member(addresses, args.struct_name, args.member_name,
                             args.member_type)



from argparse import RawTextHelpFormatter
import binascii

from fa.commands import verify_bytes

DESCRIPTION = '''reduce the result-set to those matching the given string

EXAMPLE:
    0x00000000: 01 02 03 04
    0x00000004: 30 31 32 33 -> ascii '0123'

    results = [0, 2, 4]
    -> verify-str '0123'
    results = [4]
'''


def get_parser():
    p = verify_bytes.get_parser()
    p.add_argument('--null-terminated', action='store_true')

    p.prog = 'verify-str'
    p.description = DESCRIPTION
    p.formatter_class = RawTextHelpFormatter
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    hex_str = binascii.hexlify(args.hex_str)
    hex_str += '00' if args.null_terminated else ''

    setattr(args, 'hex_str', hex_str)
    return verify_bytes.run(segments, args, addresses, **kwargs)



IDA_MODULE = False

try:
    import idc  # noqa: F401

    IDA_MODULE = True
except ImportError:
    pass


class InvalidContextException(Exception):
    pass


def get_correct_implementation(function_name, params, ida=None, unknown=None,
                               **kwargs):
    """
    Get and execute the correct implementation according to the currently
    executing context
    :param function_name: function name to be executes
    :param params: parameters to pass to function
    :param ida: IDA context implementation
    :param unknown: Unknown context implementation
    :return: The execution result from the correctly running context
    """
    if ida and IDA_MODULE:
        return ida(*params, **kwargs)
    if unknown:
        return unknown(*params, **kwargs)

    raise InvalidContextException(
        'function "{}" must be executed from a specific context'
        .format(function_name))


def verify_ida(function_name):
    if not IDA_MODULE:
        raise InvalidContextException(
            'operation "{}" must be executed from an IDA context'
            .format(function_name))


def ida_context(function):
    if IDA_MODULE:
        return function
    else:
        def invalid_context(*kwargs):
            raise InvalidContextException(
                'function "{}" must be executed from an IDA context'
                .format(function.__name__))
        return invalid_context



from argparse import RawTextHelpFormatter

from fa import utils, context

DESCRIPTION = '''expands the result-set with the occurrences of the given bytes
expression in "ida bytes syntax"

EXAMPLE:
    0x00000000: 01 02 03 04
    0x00000004: 05 06 07 08

    results = []
    -> find-bytes-ida '01 02 03 04'
    result = [0]

    -> find-bytes-ida '05 06 ?? 08'
    results = [0, 4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('find-bytes-ida',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('expression')
    return p


@context.ida_context
def find_bytes_ida(expression, segments=None):
    for address in utils.ida_find_all(expression):
        yield address


def run(segments, args, addresses, interpreter=None, **kwargs):
    return addresses + list(find_bytes_ida(args.expression))



from fa import utils


def get_parser():
    p = utils.ArgumentParserNoExit('set-const',
                                   description='define a const value')
    p.add_argument('name')
    return p


def set_const(addresses, name, interpreter):
    for ea in addresses:
        interpreter.set_const(name, ea)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return set_const(addresses, args.name, interpreter)



from fa import utils


def get_parser():
    p = utils.ArgumentParserNoExit('find',
                                   description='find another symbol defined '
                                               'in other SIG files')
    p.add_argument('name', help='symbol name')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    interpreter.find(args.name, use_cache=interpreter.implicit_use_sig_cache)

    # return an empty result-set
    return []



from argparse import RawTextHelpFormatter
from fa import utils


DESCRIPTION = '''adds a python-range to resultset

EXAMPLE:
    result = [0, 0x200]
    -> add-offset-range 0 4 8
    result = [0, 4, 8, 0x200, 0x204, 0x208]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('add-offset-range',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('start')
    p.add_argument('end')
    p.add_argument('step')
    return p


def add_offset_range(addresses, start, end, step):
    for ea in addresses:
        for i in range(start, end, step):
            yield ea + i


def run(segments, args, addresses, interpreter=None, **kwargs):
    gen = add_offset_range(addresses, eval(args.start), eval(args.end),
                           eval(args.step))
    return list(gen)



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idc
except ImportError:
    pass

DESCRIPTION = '''get operand value from given instruction

EXAMPLE #1:
    0x00000000: mov r0, r1
    0x00000004: mov r1, r2
    0x00000008: push {r4}

    results = [4]
    -> operand 1
    results = [2]  # because r2
'''


def get_parser():
    p = utils.ArgumentParserNoExit('operand',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('op', help='operand number')
    return p


@context.ida_context
def operand(addresses, op):
    for address in addresses:
        yield idc.get_operand_value(address, op)


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(operand(addresses, eval(args.op)))



from fa import utils, context
from fa.commands.locate import locate


def get_parser():
    p = utils.ArgumentParserNoExit('verify-name',
                                   description='verifies the given name '
                                               'appears in result set')
    p.add_argument('name')
    return p


@context.ida_context
@utils.yield_unique
def verify_name(addresses, name):
    ref = locate(name)
    for address in addresses:
        if ref == address:
            yield address


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(verify_name(addresses, args.name))



from fa import utils, context

try:
    import ida_funcs
except ImportError:
    pass


def get_parser():
    p = utils.ArgumentParserNoExit('make-function',
                                   description='convert into a function')
    return p


@context.ida_context
def make_function(addresses):
    for ea in addresses:
        ida_funcs.add_func(ea)
    return addresses


def run(segments, args, addresses, interpreter=None, **kwargs):
    return make_function(addresses)



from pathlib import Path

from setuptools import setup

BASE_DIR = Path(__file__).parent.resolve(strict=True)


def parse_requirements():
    reqs = []
    with open(BASE_DIR / 'requirements.txt', 'r') as fd:
        for line in fd.readlines():
            line = line.strip()
            if line:
                reqs.append(line)
    return reqs


setup(
    name='fa',
    version='0.3.0',
    description='FA Plugin',
    author='DoronZ',
    author_email='doron88@gmail.com',
    url='https://github.com/doronz88/fa',
    packages=['fa', 'fa.commands'],
    package_dir={'fa': 'fa'},
    package_data={'': ['*.png', '*'], },
    include_package_data=True,
    data_files=[(r'fa/res/icons', [r'fa/res/icons/create_sig.png',
                                   r'fa/res/icons/export.png',
                                   r'fa/res/icons/find.png',
                                   r'fa/res/icons/find_all.png',
                                   r'fa/res/icons/save.png',
                                   r'fa/res/icons/settings.png',
                                   r'fa/res/icons/suitcase.png']),
                (r'fa/commands', ['fa/commands/alias']),
                ],
    install_requires=parse_requirements(),
    python_requires='>=2.7',
)



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idc
except ImportError:
    pass

DESCRIPTION = '''goto symbol by name

EXAMPLE:
    0x00000000: main:
    0x00000000:     mov r0, r1
    0x00000004: foo:
    0x00000004:     bx lr

    results = [0, 4]
    -> locate foo
    result = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('locate',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('name')
    return p


@context.ida_context
def locate(name):
    return idc.get_name_ea_simple(name)


def run(segments, args, addresses, interpreter=None, **kwargs):
    address = locate(args.name)
    return [address] if address != idc.BADADDR else []



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import ida_typeinf
except ImportError:
    pass

DESCRIPTION = '''get function's argument assignment address

EXAMPLE:
    0x00000000: ldr r0, =dest
    0x00000004: ldr r1, =src
    0x00000008: mov r2, #4
    0x0000000c: bl memcpy

    results = [0x0c]
    -> argument 2
    results = [8]  # address of 3rd argument
'''


def get_parser():
    p = utils.ArgumentParserNoExit('argument',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('arg', help='argument number')
    return p


@context.ida_context
def argument(addresses, arg):
    for address in addresses:
        args = ida_typeinf.get_arg_addrs(address)
        if args is None:
            continue
        try:
            yield args[arg]
        except KeyError:
            continue


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(argument(addresses, eval(args.arg)))



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''align results to given base (round-up)

EXAMPLE:
    results = [0, 2, 4, 6, 8]
    -> align 4
    results = [0, 4, 4, 8, 8]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('align',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('value')
    return p


def align(addresses, value):
    return [((ea + (value - 1)) // value) * value for ea in addresses]


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(align(addresses, eval(args.value)))



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''branch unconditionally to label

EXAMPLE:
    results = []

    add 1
    -> b skip
    add 2
    label skip
    add 3

    results = [1, 3]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('b',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('label', help='label to jump to')
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    interpreter.set_pc(args.label)
    # pc is incremented by 1, after each instruction
    interpreter.dec_pc()
    return addresses



from argparse import RawTextHelpFormatter

from fa import utils, context

try:
    import idc
    import ida_search
except ImportError:
    pass

DESCRIPTION = '''expands the result-set with the occurrences of the given
immediate in "ida immediate syntax"

EXAMPLE:
    0x00000000: ldr r0, =0x1234
    0x00000004: add r0, #2 ; 0x1236

    results = []
    -> find-immediate 0x1236
    result = [4]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('find-immediate',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('expression')
    return p


@context.ida_context
def find_immediate(expression):
    if isinstance(expression, str):
        expression = eval(expression)

    ea, imm = ida_search.find_imm(0, idc.SEARCH_DOWN, expression)
    while ea != idc.BADADDR:
        yield ea
        ea, imm = idc.find_imm(ea + 1, idc.SEARCH_DOWN,
                               expression)


def run(segments, args, addresses, interpreter=None, **kwargs):
    results = list(find_immediate(args.expression))
    return addresses + results



from argparse import RawTextHelpFormatter
from fa import utils, context

try:
    import idc
except ImportError:
    pass

DESCRIPTION = '''goto function's start

EXAMPLE:
    0x00000000: push {r4-r7, lr} -> function's prolog
    ...
    0x000000f0: pop {r4-r7, pc} -> function's epilog

    results = [0xf0]
    -> function-start
    result = [0]
'''


def get_function_start(segments, ea):
    start = idc.get_func_attr(ea, idc.FUNCATTR_START)
    return start

    # TODO: consider add support locate of function heads manually


def get_parser():
    p = utils.ArgumentParserNoExit('function-start',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('cmd', nargs='*', default='', help='command')
    return p


@context.ida_context
def function_start(addresses):
    for ea in addresses:
        if ea != idc.BADADDR:
            func_start = idc.get_func_attr(ea, idc.FUNCATTR_START)
            if func_start != idc.BADADDR:
                yield func_start


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(function_start(addresses))



import time

import pkg_resources
from tkinter import ttk, Tk
from configparser import ConfigParser

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import shlex
import sys
import os

import hjson

from fa.utils import ArgumentParserNoExit

CONFIG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '..', 'config.ini')
DEFAULT_SIGNATURES_ROOT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'signatures')
COMMANDS_ROOT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'commands')


class InterpreterState:
    def __init__(self):
        self.pc = 0
        self.variables = {}
        self.labels = {}


class FaInterp:
    """
    FA Interpreter base class
    """
    __metaclass__ = ABCMeta

    def __init__(self, config_path=CONFIG_PATH):
        """
        Constructor
        :param config_path: config.ini path. used to load global settings
                            instead of setting each of the options manually
                            (signatures_root, project, ...)
        """
        self._project = None
        self._input = None
        self._segments = OrderedDict()
        self._signatures_root = DEFAULT_SIGNATURES_ROOT
        self._symbols = {}
        self._consts = {}
        self.history = []
        self.endianity = '<'
        self._config_path = config_path
        self.stack = []
        self._sig_cache = []
        self.implicit_use_sig_cache = False

        if (config_path is not None) and (os.path.exists(config_path)):
            self._signatures_root = os.path.expanduser(
                self.config_get('global', 'signatures_root'))
            self._project = self.config_get('global', 'project', None)

    def _push_stack_frame(self):
        self.stack.append(InterpreterState())

    def _pop_stack_frame(self):
        self.stack.pop()

    def set_labels(self, labels):
        self.stack[-1].labels = labels

    def get_pc(self):
        return self.stack[-1].pc

    def set_pc(self, pc):
        if isinstance(pc, int):
            self.stack[-1].pc = pc
        elif isinstance(pc, str):
            self.stack[-1].pc = self.stack[-1].labels[pc]
        else:
            raise KeyError('invalid pc: {}'.format(pc))

    def dec_pc(self):
        self.stack[-1].pc -= 1

    def inc_pc(self):
        self.stack[-1].pc += 1

    def set_variable(self, name, value):
        self.stack[-1].variables[name] = value

    def get_variable(self, name):
        return self.stack[-1].variables[name]

    def get_all_variables(self):
        return self.stack[-1].variables

    @abstractmethod
    def set_input(self, input_):
        """
        Set file input
        :param input_: file to work on
        :return:
        """
        pass

    def config_get(self, section, key, default=None):
        """
        Read configuration setting. This is loaded from INI config file.
        :param section: section name
        :param key: key name
        :param default: default value, if key doesn't exist inside section
        :return: the value in the specified section-key
        """
        if not os.path.exists(self._config_path):
            return default

        config = ConfigParser()

        with open(self._config_path) as f:
            config.read_file(f)

        if not config.has_section(section) or \
                not config.has_option(section, key):
            return default

        return config.get(section, key)

    def config_set(self, section, key, value):
        """
        Write configuration setting. This is saved into INI config file
        :param section: section name
        :param key: key name
        :param value: value to set
        :return: None
        """
        config = ConfigParser()

        if sys.version[0] == '2':
            section = section.decode('utf8')
            key = key.decode('utf8')
            value = value.decode('utf8')

        if os.path.exists(self._config_path):
            config.read(self._config_path)

        if not config.has_section(section):
            config.add_section(section)

        config.set(section, key, value)

        with open(self._config_path, 'w') as f:
            config.write(f)

    def set_signatures_root(self, path, save=False):
        """
        Change signatures root path (where the projects are searched).
        :param path: signatures root path (where the projects are searched).
        :param save: should save into configuration file?
        :return: None
        """
        self.log('signatures root: {}'.format(path))
        self._signatures_root = path

        if save:
            self.config_set('global', 'signatures_root', path)

    def verify_project(self):
        """
        Throws IOError if no project has been selected or points into an
        invalid path
        :return: None
        """
        if self._project is None:
            raise IOError('No project has been selected')

        if not os.path.exists(os.path.join(self._signatures_root,
                                           self._project)):
            raise IOError("Selected project's path doesn't exist.\n"
                          "Please re-select)")

    def set_project(self, project, save=True):
        """
        Set currently active project (where SIG files are placed)
        :param project: project name
        :param save: should save this setting into configuration file?
        :return: None
        """
        self._project = project
        self.log('project set: {}'.format(project))

        self.set_signatures_root(self._signatures_root, save=save)
        if save:
            self.config_set('global', 'project', project)

    def symbols(self):
        """
        Run find for all SIG files in currently active project
        :return: dictionary of found symbols
        """
        try:
            self._sig_cache = []
            self.implicit_use_sig_cache = True
            self.get_python_symbols()

            for sig in self.get_json_signatures():
                self.find(sig['name'], use_cache=True)
        finally:
            self._sig_cache = []
            self.implicit_use_sig_cache = False

        return self._symbols

    def interactive_set_project(self):
        """
        Show GUI for selecting a project from signatures_root
        :return: None
        """
        app = Tk()
        # app.geometry('200x30')

        label = ttk.Label(app,
                          text="Choose current project")
        label.grid(column=0, row=0)

        combo = ttk.Combobox(app,
                             values=self.list_projects())
        combo.grid(column=0, row=1)

        def combobox_change_project(event):
            self.set_project(combo.get())

        combo.bind("<<ComboboxSelected>>", combobox_change_project)

        app.mainloop()

    def list_projects(self):
        """
        Get a list of all available projects in signatures_root
        :return: list of all available projects in signatures_root
        """
        projects = []
        for root, dirs, files in os.walk(self._signatures_root):
            projects += \
                [os.path.relpath(os.path.join(root, filename),
                                 self._signatures_root) for filename in dirs]
        return [str(p) for p in projects if p[0] != '.']

    @staticmethod
    def log(message):
        """
        Log message
        :param message:
        :return:
        """
        for line in str(message).splitlines():
            print('FA> {}'.format(line))

    @abstractmethod
    def reload_segments(self):
        """
        Reload memory segments
        :return:
        """
        pass

    @staticmethod
    def get_module(name, filename):
        """
        Load a python module by filename
        :param name: module name
        :param filename: module filename
        :return: loaded python module
        """
        if not os.path.exists(filename):
            raise NotImplementedError("no such filename: {}".format(filename))

        if sys.version == '3':
            # TODO: support python 3.0-3.4
            import importlib.util
            spec = importlib.util.spec_from_file_location(name, filename)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        else:
            import imp
            module = imp.load_source(name, filename)

        return module

    @staticmethod
    def get_command(command):
        """
        Get fa command as a loaded python-module
        :param command: command name
        :return: command's python-module
        """
        filename = os.path.join(COMMANDS_ROOT, "{}.py".format(command))
        return FaInterp.get_module(command, filename)

    def run_command(self, command, addresses):
        """
        Run fa command with given address list and output the result
        :param command: fa command name
        :param addresses: input address list
        :return: output address list
        """
        args = ''
        if ' ' in command:
            command, args = command.split(' ', 1)
            args = shlex.split(args)

        command = command.replace('-', '_')

        module = self.get_command(command)
        p = module.get_parser()
        args = p.parse_args(args)
        return module.run(self._segments, args, addresses,
                          interpreter=self)

    def get_alias(self):
        """
        Get dictionary of all defined aliases globally and by project.
        Project aliases loaded last so are considered stronger.
        :return: dictionary of all fa command aliases
        """
        retval = {}

        alias_res_path = os.path.join('commands', 'alias')
        alias_filename = pkg_resources.resource_filename('fa', alias_res_path)
        with open(alias_filename) as f:
            for line in f.readlines():
                line = line.strip()
                k, v = line.split('=')
                retval[k.strip()] = v.strip()

        if self._project:
            # include also project alias
            project_root = os.path.join(self._signatures_root, self._project)
            project_alias_filename = os.path.join(project_root, 'alias')
            if os.path.exists(project_alias_filename):
                with open(project_alias_filename) as f:
                    for line in f.readlines():
                        line = line.strip()
                        k, v = line.split('=')
                        retval[k.strip()] = v.strip()

        return retval

    def save_signature(self, filename):
        """
        Save given signature object (by dictionary) into active project
        as a new SIG file. If symbol name already exists, then create another
        file (never overwrites).
        :param filename: Dictionary of signature object
        :return: None
        """
        with open(filename) as f:
            sig = hjson.load(f)
            f.seek(0)
            sig_text = f.read()

        filename = os.path.join(
            self._signatures_root,
            self._project,
            sig['name'] + '.sig')
        i = 1
        while os.path.exists(filename):
            filename = os.path.join(self._signatures_root, self._project,
                                    sig['name'] + '.{}.sig'.format(i))
            i += 1

        with open(filename, 'w') as f:
            f.write(sig_text)

    def _get_labeled_instructions(self, instructions):
        labels = {}
        processed_instructions = []

        label_parser = ArgumentParserNoExit('label')
        label_parser.add_argument('name')

        alias_items = self.get_alias().items()

        pc = 0
        for line in instructions:
            line = line.strip()

            if len(line) == 0:
                continue

            if line.startswith('#'):
                # treat as comment
                continue

            if line.startswith('label '):
                args = label_parser.parse_args(shlex.split(line)[1:])
                labels[args.name] = pc
                continue

            for k, v in alias_items:
                # handle aliases
                if line.startswith(k):
                    line = line.replace(k, v)

            processed_instructions.append(line)
            pc += 1

        return labels, processed_instructions

    def find_from_instructions_list(self, instructions, addresses=None):
        """
        Run the given instruction list and output the result
        :param instructions: instruction list
        :param addresses: input address list (if any)
        :return: output address list
        """
        if addresses is None:
            addresses = []

        self._push_stack_frame()

        labels, instructions = self._get_labeled_instructions(instructions)

        self.set_labels(labels)

        n = len(instructions)

        while self.get_pc() < n:
            line = instructions[self.get_pc()]

            new_addresses = []
            try:
                new_addresses = self.run_command(line, addresses)
            except ImportError as m:
                FaInterp.log('failed to run: {}. error: {}'
                             .format(line, str(m)))

            addresses = new_addresses
            self.inc_pc()

        self._pop_stack_frame()
        return addresses

    def find_from_sig_json(self, signature_json, decremental=False):
        """
        Find a signature from a signature JSON data.
        :param dict signature_json: Data of signature's JSON.
        :param bool decremental:
        :return: Addresses of matching signatures.
        :rtype: result list of last returns instruction
        """
        self.log('interpreting SIG for: {}'.format(signature_json['name']))
        start = time.time()
        retval = self.find_from_instructions_list(
            signature_json['instructions'])
        self.log('interpretation took: {}s'.format(time.time() - start))
        return retval

    def find_from_sig_path(self, signature_path, decremental=False):
        """
        Find a signature from a signature file path.
        :param str signature_path: Path to a signature file.
        :param bool decremental:
        :return: Addresses of matching signatures.
        :rtype: result list of last returns instruction
        """
        local_path = os.path.join(
            self._signatures_root, self._project, signature_path)
        if os.path.exists(local_path):
            # prefer local signatures, then external
            signature_path = local_path

        with open(signature_path) as f:
            sig = hjson.load(f)
        return self.find_from_sig_json(sig, decremental)

    def get_python_symbols(self, file_name=None):
        """
        Run all python scripts found in currently active project and return
        the dictionary of all found symbols
        :param file_name: optional, specify which python script to execute
                          inside the currently active project
        :return: dictionary of all found symbols
        """
        project_root = os.path.join(self._signatures_root, self._project)
        sys.path.append(project_root)

        for root, dirs, files in os.walk(project_root):
            for filename in sorted(files):
                if not filename.lower().endswith('.py'):
                    continue

                if not file_name or file_name == filename:
                    name = os.path.splitext(filename)[0]
                    filename = os.path.join(root, filename)
                    m = FaInterp.get_module(name, filename)
                    if not hasattr(m, 'run'):
                        self.log('skipping: {}'.format(filename))
                    else:
                        m.run(self)

    def get_json_signatures(self, symbol_name=None):
        """
        Get a list of all json SIG objects in currently active project.
        :param symbol_name: optional, select a specific SIG file by symbol name
        :return: list of all json SIG objects in currently active project.
        """
        signatures = []
        project_root = os.path.join(self._signatures_root, self._project)

        for root, dirs, files in os.walk(project_root):
            for filename in sorted(files):
                if not filename.lower().endswith('.sig'):
                    continue

                filename = os.path.join(project_root, filename)

                with open(filename) as f:
                    try:
                        signature = hjson.load(f)
                    except ValueError as e:
                        self.log('error in json: {}'.format(filename))
                        raise e

                if (symbol_name is None) or (signature['name'] == symbol_name):
                    signatures.append(signature)

        return signatures

    def set_const(self, name, value):
        self._consts[name] = value

    def set_symbol(self, symbol_name, value):
        self._symbols[symbol_name] = value

    def get_consts(self):
        return self._consts

    def find(self, symbol_name, use_cache=False):
        """
        Find symbol by its name in the SIG file
        :param symbol_name: symbol name
        :return: list of matches for the given symbol
        """
        if use_cache and symbol_name in self._sig_cache:
            return

        results = []
        signatures = self.get_json_signatures(symbol_name)
        if len(signatures) == 0:
            raise NotImplementedError('no signature found for: {}'
                                      .format(symbol_name))

        for sig in signatures:
            sig_results = self.find_from_sig_json(sig)

            if isinstance(sig_results, dict):
                if symbol_name in sig_results:
                    results += sig_results[symbol_name]
            else:
                results += sig_results

        if use_cache and symbol_name not in self._sig_cache:
            self._sig_cache.append(symbol_name)

        return list(set(results))



from argparse import RawTextHelpFormatter
from fa import utils
from fa import context

try:
    import idc
except ImportError:
    pass

DESCRIPTION = '''goto function's end

EXAMPLE:
    0x00000000: push {r4-r7, lr} -> function's prolog
    ...
    0x000000f0: push {r4-r7, pc} -> function's epilog

    results = [0]
    -> function-end
    result = [0xf0]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('function-end',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    return p


@context.ida_context
def function_end(addresses):
    for ea in addresses:
        if ea != idc.BADADDR:
            func_end = idc.get_func_attr(ea, idc.FUNCATTR_END)
            if func_end != idc.BADADDR:
                yield func_end


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(function_end(addresses))



from argparse import RawTextHelpFormatter

try:
    import idc
except ImportError:
    pass

from fa import utils, context

DESCRIPTION = '''reduce the result-set to those in the given segment name

EXAMPLE:
    .text:0x00000000 01 02 03 04
    .text:0x00000004 30 31 32 33

    .data:0x00000200 01 02 03 04
    .data:0x00000204 30 31 32 33

    results = [0, 0x200]
    -> verify-segment .data
    results = [0x200]
'''


@context.ida_context
def verify_segment(addresses, segment_name):
    for ea in addresses:
        if segment_name == idc.get_segm_name(ea):
            yield ea


def get_parser():
    p = utils.ArgumentParserNoExit()
    p.add_argument('name', help='segment name')

    p.prog = 'verify-segment'
    p.description = DESCRIPTION
    p.formatter_class = RawTextHelpFormatter
    return p


def run(segments, args, addresses, interpreter=None, **kwargs):
    return list(verify_segment(addresses, args.name))



from argparse import RawTextHelpFormatter
from fa import utils

DESCRIPTION = '''peek a single result from the result-set (zero-based)

EXAMPLE:
    results = [0, 4, 8, 12]
    -> single 2
    result = [8]
'''


def get_parser():
    p = utils.ArgumentParserNoExit('single',
                                   description=DESCRIPTION,
                                   formatter_class=RawTextHelpFormatter)
    p.add_argument('index', default='0', help='result index')
    return p


def single(addresses, index):
    if index + 1 > len(addresses):
        return []
    else:
        return [addresses[index]]


def run(segments, args, addresses, interpreter=None, **kwargs):
    return single(addresses, eval(args.index))
